/*
 * QEstEidUtil
 *
 * Copyright (C) 2009-2011 Jargo Kõster <jargo@innovaatik.ee>
 * Copyright (C) 2009-2011 Raul Metsma <raul@innovaatik.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "CertUpdate.h"
#include "DiagnosticsDialog.h"
#include "QSmartCard.h"
#include "SettingsDialog.h"
#include "XmlReader.h"
#ifdef Q_OS_WIN
#include "CertStore.h"
#endif

#include <common/AboutWidget.h>
#include <common/CertificateWidget.h>
#include <common/Common.h>
#include <common/MessageBox.h>
#include <common/Settings.h>
#include <common/SslCertificate.h>
#include <common/sslConnect.h>

#include <QDate>
#include <QDesktopServices>
#include <QFileDialog>
#include <QMenuBar>
#include <QTextStream>
#include <QTranslator>
#include <QUrl>

class MainWindowPrivate: public Ui::MainWindow
{
	Q_DECLARE_TR_FUNCTIONS(MainWindow)
	Q_DECLARE_PUBLIC(::MainWindow)
public:
	MainWindowPrivate( ::MainWindow *q ): q_ptr(q), menu(0), loading(0), savePicture(0), b(0) {}

	void clearPins();
	void hideLoading();
	QByteArray sendRequest( SSLConnect::RequestType type, const QString &param = QString() );
	void showLoading( const QString &text );
	void showWarning( const QString &msg, const QString &details = QString() );
	bool validateCardError( QSmartCardData::PinType type, int flags, QSmartCard::ErrorType err );
	bool validatePin( QSmartCardData::PinType type, bool puk, const QString &old, const QString &pin, const QString &pin2 );

	::MainWindow *q_ptr;
	QTranslator *appTranslator, *qtTranslator, *commonTranslator;
	QMenuBar *bar;
	QMenu *menu;
	QAction *about, *pref, *close;
	QSmartCard *smartcard;
	QLabel *loading, *savePicture;
	QButtonGroup *b;
};



void MainWindowPrivate::clearPins()
{
	changePin1Repeat->clear();
	changePin1New->clear();
	changePin1Validate->clear();
	changePin2Repeat->clear();
	changePin2New->clear();
	changePin2Validate->clear();
	changePukRepeat->clear();
	changePukNew->clear();
	changePukValidate->clear();
	switch( dataWidget->currentIndex() )
	{
	case ::MainWindow::PagePin1Pin: changePin1Validate->setFocus(); break;
	case ::MainWindow::PagePin2Pin: changePin2Validate->setFocus(); break;
	case ::MainWindow::PagePuk: changePukValidate->setFocus(); break;
	default: break;
	}
}

void MainWindowPrivate::hideLoading()
{
	loading->hide();
	loading->parentWidget()->setEnabled( true );
}

QByteArray MainWindowPrivate::sendRequest( SSLConnect::RequestType type, const QString &param )
{
	Q_Q(::MainWindow);
	switch( type )
	{
	case SSLConnect::ActivateEmails: showLoading(  tr("Activating e-mail settings") ); break;
	case SSLConnect::EmailInfo: showLoading( tr("Loading e-mail settings") ); break;
	case SSLConnect::MobileInfo: showLoading( tr("Loading Mobiil-ID settings") ); break;
	case SSLConnect::PictureInfo: showLoading( tr("Loading picture") ); break;
	default: showLoading( tr( "Loading data" ) ); break;
	}

	if( !validateCardError( QSmartCardData::Pin1Type, type, smartcard->login( QSmartCardData::Pin1Type ) ) )
		return QByteArray();

	SSLConnect ssl;
	ssl.setToken( smartcard->data().authCert(), smartcard->key() );
	QByteArray buffer = ssl.getUrl( type, param );
	smartcard->logout();
	q->updateData();
	if( !ssl.errorString().isEmpty() )
	{
		switch( type )
		{
		case SSLConnect::ActivateEmails: showWarning( tr("Failed activating e-mail forwards."), ssl.errorString() ); break;
		case SSLConnect::EmailInfo: showWarning( tr("Failed loading e-mail settings."), ssl.errorString() ); break;
		case SSLConnect::MobileInfo: showWarning( tr("Failed loading Mobiil-ID settings."), ssl.errorString() ); break;
		case SSLConnect::PictureInfo: showWarning( tr("Loading picture failed."), ssl.errorString() ); break;
		default: showWarning( tr("Failed to load data"), ssl.errorString() ); break;
		}
		return QByteArray();
	}
	return buffer;
}

void MainWindowPrivate::showLoading( const QString &text )
{
	Q_Q(::MainWindow);
	loading->setFixedSize( text.size() > 20 ? 300 : 250, 100 );
	loading->move( q->width()/2 - loading->width()/2, 305 );
	loading->setText( text );
	loading->show();
	loading->parentWidget()->setDisabled( true );
}

void MainWindowPrivate::showWarning( const QString &msg, const QString &details )
{
	DMessageBox d( QMessageBox::Warning, tr("ID-card utility"), msg, QMessageBox::Close, qApp->activeWindow() );
	d.setWindowModality( Qt::WindowModal );
	if( !details.isEmpty() )
	{
		d.addButton( QMessageBox::Help );
		d.setDetailedText( details );
	}
	if( d.exec() == QMessageBox::Help )
		return;//Common::showHelp( msg, err );
}

bool MainWindowPrivate::validateCardError( QSmartCardData::PinType type, int flags, QSmartCard::ErrorType err )
{
	Q_Q(::MainWindow);
	q->updateData();
	QSmartCardData::PinType t = flags == 1025 ? QSmartCardData::PukType : type;
	switch( err )
	{
	case QSmartCard::NoError: return true;
	case QSmartCard::CancelError: break;
	case QSmartCard::BlockedError:
		showWarning( tr("%1 blocked").arg( QSmartCardData::typeString( t ) ) ); break;
	case QSmartCard::DifferentError:
		showWarning( tr("New %1 codes doesn't match").arg( QSmartCardData::typeString( type ) ) ); break;
	case QSmartCard::LenghtError:
		switch( type )
		{
		case QSmartCardData::Pin1Type: showWarning( tr("PIN1 length has to be between 4 and 12") ); break;
		case QSmartCardData::Pin2Type: showWarning( tr("PIN2 length has to be between 5 and 12") ); break;
		case QSmartCardData::PukType: showWarning( tr("PUK length has to be between 8 and 12") ); break;
		}
		break;
	case QSmartCard::ValidateError:
		showWarning( tr("Wrong %1 code. You can try %n more time(s).", "",
			smartcard->data().retryCount( t ) ).arg( QSmartCardData::typeString( t ) ) );
		break;
	default:
		switch( flags )
		{
		case SSLConnect::ActivateEmails: showWarning( tr("Failed activating e-mail forwards.") ); break;
		case SSLConnect::EmailInfo: showWarning( tr("Failed loading e-mail settings.") ); break;
		case SSLConnect::MobileInfo: showWarning( tr("Failed loading Mobiil-ID settings.") ); break;
		case SSLConnect::PictureInfo: showWarning( tr("Loading picture failed.") ); break;
		default:
			showWarning( tr("Changing %1 failed").arg( QSmartCardData::typeString( type ) ) ); break;
		}
		break;
	}
	return false;
}

bool MainWindowPrivate::validatePin( QSmartCardData::PinType type, bool puk,
	const QString &old, const QString &pin, const QString &pin2 )
{
	QString name = QSmartCardData::typeString( type );
	if( old.isEmpty() )
	{ showWarning( puk ? tr("Enter PUK code") : tr("Enter current %1 code").arg( name ) ); return false; }
	if( pin.isEmpty() )
	{ showWarning( tr("Enter new %1 code").arg( name ) ); return false; }
	if( pin2.isEmpty() )
	{ showWarning( tr("Retry your new %1 code").arg( name ) ); return false; }
	if( pin == old )
	{ showWarning( tr("Old and new %1 has to be different!").arg( name ) ); return false; }
	if( pin != pin2 )
	{ showWarning( tr("New %1 codes doesn't match").arg( name ) ); return false; }

	switch( type )
	{
	case QSmartCardData::Pin1Type:
		if( pin.size() < 4 || pin.size() > 12 )
		{ showWarning( tr("PIN1 length has to be between 4 and 12") ); return false; }
		break;
	case QSmartCardData::Pin2Type:
		if( pin.size() < 5 || pin.size() > 12 )
		{ showWarning( tr("PIN2 length has to be between 5 and 12") ); return false; }
		break;
	case QSmartCardData::PukType:
		if( pin.size() < 8 || pin.size() > 12 )
		{ showWarning( tr("PUK length has to be between 8 and 12") ); return false; }
		break;
	}

	QSmartCardData t = smartcard->data();
	if( t.cardType() == QSmartCardData::Tempel )
		return true;

	QDate date = t.data( QSmartCardData::BirthDate ).toDate();
	if( !date.isNull() &&
		(pin.contains( date.toString( "yyyy" ) ) ||
		pin.contains( date.toString( "ddMM" ) ) ||
		pin.contains( date.toString( "MMdd" ) )) )
	{
		showWarning( tr("%1 have to be different than date of birth or year of birth").arg( name ) );
		return false;
	}
	return true;
}



MainWindow::MainWindow( QWidget *parent )
:	QWidget( parent )
,	d( new MainWindowPrivate(this) )
{
	setWindowFlags( Qt::Window|Qt::CustomizeWindowHint|Qt::WindowMinimizeButtonHint|Qt::WindowCloseButtonHint );
	setFixedSize( 585, 535 );
	d->setupUi( this );

	d->loading = new QLabel( d->centralwidget );
	d->loading->setObjectName( "loading" );
	d->loading->setAlignment( Qt::AlignCenter );
	d->loading->setWordWrap( true );
	d->savePicture = new QLabel( d->loadPicture );
	d->savePicture->setObjectName( "savePicture" );
	d->savePicture->setAlignment( Qt::AlignCenter );
	QVBoxLayout *l = new QVBoxLayout( d->loadPicture );
	l->setMargin( 0 );
	l->addSpacing( 0 );
	l->addWidget( d->savePicture, 0, Qt::AlignBottom );

	QRegExpValidator *validator = new QRegExpValidator( QRegExp( "\\d*" ), this );
	d->changePin1Validate->setValidator( validator );
	d->changePin1New->setValidator( validator );
	d->changePin1Repeat->setValidator( validator );
	d->changePin2Validate->setValidator( validator );
	d->changePin2New->setValidator( validator );
	d->changePin2Repeat->setValidator( validator );
	d->changePukValidate->setValidator( validator );
	d->changePukNew->setValidator( validator );
	d->changePukRepeat->setValidator( validator );

	d->cardInfo->hide();
	d->cards->hide();
	d->cards->hack();
	d->languages->hack();
	d->languages->setItemData( 0, "et" );
	d->languages->setItemData( 1, "en" );
	d->languages->setItemData( 2, "ru" );
	lang = "et";
	d->frameRegNr->hide();

	d->b = new QButtonGroup( this );
	// Main buttons
	d->b->addButton( d->buttonCert, PageCert );
	d->b->addButton( d->buttonEmail, PageEmail );
	d->b->addButton( d->buttonMobile, PageMobile );
	d->b->addButton( d->buttonPuk, PagePukInfo );
	// Cert page buttons
	d->b->addButton( d->authCert, PageCertAuthView );
	d->b->addButton( d->signCert, PageCertSignView );
	d->b->addButton( d->updateCert, PageCertUpdate );
	d->b->addButton( d->authChangePin, PagePin1Pin );
	d->b->addButton( d->authRevoke, PagePin1Unblock );
	d->b->addButton( d->signChangePin, PagePin2Pin );
	d->b->addButton( d->signRevoke, PagePin2Unblock );
	// puk info buttons
	d->b->addButton( d->pukChange, PagePuk );
	// Email buttons
	d->b->addButton( d->checkEmail, PageEmailStatus );
	d->b->addButton( d->activateEmail, PageEmailActivate );
	// mobile buttons
	d->b->addButton( d->checkMobile, PageMobileStatus );
	d->b->addButton( d->mobileActivate, PageMobileActivate );
	// pin1 buttons
	d->b->addButton( d->changePin1Cancel, PageCert );
	d->b->addButton( d->changePin1Change, PagePin1ChangePin );
	d->b->addButton( d->changePin1PinpadCancel, PageCert );
	d->b->addButton( d->changePin1PinpadChange, PagePin1ChangePin );
	//pin2 buttons
	d->b->addButton( d->changePin2Cancel, PageCert );
	d->b->addButton( d->changePin2Change, PagePin2ChangePin );
	d->b->addButton( d->changePin2PinpadCancel, PageCert );
	d->b->addButton( d->changePin2PinpadChange, PagePin2ChangePin );
	//puk buttons
	d->b->addButton( d->changePukCancel, PageCert );
	d->b->addButton( d->changePukChange, PagePukChange );
	d->b->addButton( d->changePukPinpadCancel, PageCert );
	d->b->addButton( d->changePukPinpadChange, PagePukChange );

	connect( d->b, SIGNAL(buttonClicked(int)), SLOT(setDataPage(int)) );
	connect( d->savePicture, SIGNAL(linkActivated(QString)), SLOT(parseLink(QString)) );

	qApp->installTranslator( d->appTranslator = new QTranslator( this ) );
	qApp->installTranslator( d->qtTranslator = new QTranslator( this ) );
	qApp->installTranslator( d->commonTranslator = new QTranslator( this ) );

#if defined(Q_OS_LINUX)
	d->headerLine1->hide();
	d->headerSettings->hide();
#elif defined(Q_OS_MAC)
	d->bar = new QMenuBar;
	d->menu = d->bar->addMenu( tr("&File") );
	d->pref = d->menu->addAction( tr("Settings"), this, SLOT(showSettings()) );
	d->about = d->menu->addAction( tr("Info"), this, SLOT(showAbout()) );
	d->close = d->menu->addAction( tr("Close"), qApp, SLOT(quit()) );
	d->pref->setMenuRole( QAction::PreferencesRole );
	d->about->setMenuRole( QAction::AboutRole );
	d->close->setShortcut( Qt::CTRL + Qt::Key_W );
#endif

	d->smartcard = new QSmartCard( this );
	connect( d->smartcard, SIGNAL(dataChanged()), SLOT(updateData()) );
	d->smartcard->start();
	connect( d->cards, SIGNAL(activated(QString)), d->smartcard, SLOT(selectCard(QString)), Qt::QueuedConnection );

	setDataPage( PageEmpty );
	int index = d->languages->findData( Settings::language() );
	if( index != -1 )
	{
		d->languages->setCurrentIndex( index );
		on_languages_activated( index );
	}
}

MainWindow::~MainWindow()
{
#ifdef Q_OS_MAC
	delete d->bar;
#endif
}

void MainWindow::on_languages_activated( int index )
{
	lang = d->languages->itemData( index ).toString();
	if( lang == "en" ) QLocale::setDefault( QLocale( QLocale::English, QLocale::UnitedKingdom ) );
	else if( lang == "ru" ) QLocale::setDefault( QLocale( QLocale::Russian, QLocale::RussianFederation ) );
	else QLocale::setDefault( QLocale( QLocale::Estonian, QLocale::Estonia ) );

	d->appTranslator->load( ":/translations/" + lang );
	d->qtTranslator->load( ":/translations/qt_" + lang );
	d->commonTranslator->load( ":/translations/common_" + lang );
	d->retranslateUi( this );
	if( !d->loadPicture->text().isEmpty() )
		d->loadPicture->setText( tr("<a href=\"loadPicture\">Load picture</a>") );
	d->savePicture->setText( tr("<a href=\"savePicture\">save</a>") );
	updateData();
#ifdef Q_OS_MAC
	if( d->menu )
	{
		d->menu->setTitle( tr("&File") );
		d->pref->setText( tr("Settings") );
		d->about->setText( tr("Info") );
		d->close->setText( tr("Close") );
	}
#endif
	Settings().setValue( "Main/Language", lang );
}

void MainWindow::parseLink( const QString &link )
{
	if( link == "loadPicture" )
	{
		QByteArray buffer = d->sendRequest( SSLConnect::PictureInfo );
		d->hideLoading();
		if( buffer.isEmpty() )
			return;

		QPixmap pix;
		d->savePicture->setVisible( pix.loadFromData( buffer ) );
		d->loadPicture->setProperty( "PICTURE", pix );
		if( !d->savePicture->isVisible() )
		{
			XmlReader xml( buffer );
			QString error;
			xml.readEmailStatus( error );
			if( !error.isEmpty() )
				showWarning( XmlReader::emailErr( error.toUInt() ) );
			return;
		}
		d->loadPicture->setPixmap( pix.scaled( 90, 120, Qt::IgnoreAspectRatio, Qt::SmoothTransformation ) );
		d->savePicture->show();
	}
	else if( link == "savePicture" )
	{
		QString file = QDesktopServices::storageLocation( QDesktopServices::PicturesLocation );
		file += "/" + d->smartcard->data().card();
		file = QFileDialog::getSaveFileName( this, tr("Save picture"), file,
			tr("JPEG (*.jpg *.jpeg);;PNG (*.png);;TIFF (*.tif *.tiff);;24-bit Bitmap (*.bmp)") );
		if( file.isEmpty() )
			return;
		QStringList exts = QStringList() << "png" << "jpg" << "jpeg" << "tiff" << "bmp";
		if( !exts.contains( QFileInfo( file ).suffix(), Qt::CaseInsensitive ) )
			file.append( ".jpg" );
		QPixmap pix = d->loadPicture->property( "PICTURE" ).value<QPixmap>();
		if( !pix.save( file ) )
			showWarning( tr("Saving picture failed!") );
	}
	else if( link == "changePIN1PUK" )
	{
		setDataPage( PagePin1Puk );
	}
	else if( link == "changePIN2PUK" )
	{
		setDataPage( PagePin2Puk );
	}
}

void MainWindow::raiseAndRead()
{
	raise();
	showNormal();
	activateWindow();
	d->smartcard->reload();
}

void MainWindow::setDataPage( int index )
{
	QSmartCardData t = d->smartcard->data();
	ButtonTypes page = ButtonTypes(index & 0x0f);
	d->dataWidget->setCurrentIndex( t.authCert().isNull() ? PageEmpty : page );
	switch( page )
	{
	case PageCert: d->buttonCert->setChecked( true ); break;
	case PageEmail: d->buttonEmail->setChecked( true ); break;
	case PageMobile: d->buttonMobile->setChecked( true ); break;
	case PagePukInfo: d->buttonPuk->setChecked( true ); break;
	default: if( QAbstractButton *b = d->b->checkedButton() ) b->setChecked( false ); break;
	}
	if( t.authCert().isNull() )
		return;

	switch( index )
	{
	case PageCert:
		d->clearPins();
		break;
	case PageCertAuthView:
		CertificateDialog( d->smartcard->data().authCert(), this ).exec();
		break;
	case PageCertSignView:
		CertificateDialog( d->smartcard->data().signCert(), this ).exec();
		break;
	case PageCertUpdate:
	{
		QMessageBox::StandardButton b = QMessageBox::warning( this, tr("Certificate update"),
			tr("For updating certificates please close all programs which are interacting with smartcard "
			   "(qdigidocclient, qdigidoccrypto, Firefox, Safari, Internet Explorer...)<br />"
			   "After updating certificates it will no longer be possible to decrypt documents that were encrypted with the old certificate.<br />"
			   "Do you want to continue?"),
			QMessageBox::Yes|QMessageBox::No, QMessageBox::No );
		if( b == QMessageBox::No )
			break;
		try
		{
			CertUpdate c( d->smartcard, this );
			if( !c.checkUpdateAllowed() )
				break;
			d->showLoading( tr("Updating certificates") );
			c.startUpdate();
			QMessageBox::information( this, tr("Certificate update"), tr("Updating certificates successful") );
			d->smartcard->reload();
		}
		catch( const std::runtime_error &e )
		{
			QMessageBox::warning( this, tr("Certificate update"), tr("Certificate update failed: %1").arg( QString::fromUtf8( e.what() ) ), QMessageBox::Ok );
			break;
		}
#if defined(Q_OS_MAC)
		QMessageBox::warning( this, tr("Certificate update"), tr("Remove card from reader to clean TokenCache. It requires user password") );
		if( Common::runPrivileged( "/bin/sh", QStringList() << "-c" << "/bin/rm -rf /var/db/TokenCache/tokens/com.apple.tokend.opensc*" ) )
			QMessageBox::information( this, tr("Certificate update"), tr("TokenCache succesfully cleaned") );
		else
			QMessageBox::warning( this, tr("Certificate update"), tr("TokenCache cleanup failed") );
#elif defined(Q_OS_WIN)
		CertStore s;
		s.remove( d->smartcard->data().authCert() );
		s.remove( d->smartcard->data().signCert() );
		if( true )
			QMessageBox::information( this, tr("Certificate update"), tr("Old certificates succesfully removed from certificate store") );
		else
			QMessageBox::warning( this, tr("Certificate update"), tr("Failed to remove old certificates from certificate store") );
#endif
		break;
	}
	case PageEmail:
		d->emailStatus->clear();
		d->activateEmailFrame->hide();
		d->checkEmailFrame->show();
		break;
	case PageEmailActivate:
	{
		if( d->activateEmailAddress->text().isEmpty() || d->activateEmailAddress->text().indexOf( "@" ) == -1 )
		{
			d->showWarning( tr("E-mail address missing or invalid!") );
			break;
		}
		QByteArray buffer = d->sendRequest( SSLConnect::ActivateEmails, d->activateEmailAddress->text() );
		if( buffer.isEmpty() )
			break;
		XmlReader xml( buffer );
		QString error;
		xml.readEmailStatus( error );
		d->emailStatus->setText( XmlReader::emailErr( error.toUInt() ) );
		d->activateEmailFrame->hide();
		break;
	}
	case PageEmailStatus:
	{
		d->emailStatus->clear();
		QByteArray buffer = d->sendRequest( SSLConnect::EmailInfo );
		if( buffer.isEmpty() )
			break;
		XmlReader xml( buffer );
		QString error;
		QStringList emails = xml.readEmailStatus( error );
		quint8 code = error.toUInt();
		if( emails.isEmpty() || code > 0 )
		{
			code = code ? code : 20;
			d->emailStatus->setText( XmlReader::emailErr( code ) );
			if( code == 20 )
			{
				d->activateEmailFrame->show();
				d->activateEmailAddress->clear();
				d->activateEmailAddress->setFocus();
			}
		}
		else
			d->emailStatus->setText( emails.join( "<br />" ) );
		d->checkEmailFrame->hide();
		break;
	}
	case PageMobile:
		d->mobileStatus->clear();
		d->checkMobileFrame->show();
		d->mobileActivateFrame->hide();
		break;
	case PageMobileStatus:
	{
		d->mobileStatus->clear();
		QByteArray buffer = d->sendRequest( SSLConnect::MobileInfo );
		if( buffer.isEmpty() )
			break;
		XmlReader xml( buffer );
		QString error;
		QHash<QString,QString> mobile = xml.readMobileStatus( error );
		quint8 code = error.toUInt();
		if( code )
		{
			showWarning( XmlReader::mobileErr( code ) );
			break;
		}
		QString text;
		QTextStream s( &text );
		s << mobile.value( "MSISDN" ) << "<br />";
		s << tr("Mobile operator") << ": " << mobile.value( "Operator" ) << "<br />";
		s << tr("Mobile status") << ": <span style=\"color: "
		  << (mobile.value( "Status" ) == "Active" ? "#509b00" : "#e80303") << "\">"
		  << XmlReader::mobileStatus( mobile.value( "Status" ) ) << "</span>" << "<br />";
		s << tr("Certificates are valid till") << ": " << mobile.value( "MIDCertsValidTo" );
		d->mobileStatus->setText( text );
		d->checkMobileFrame->hide();
		d->mobileActivateFrame->setVisible( mobile.value( "Status" ) != "Active" );
		break;
	}
	case PageMobileActivate:
		QDesktopServices::openUrl( tr("http://politsei.ee/en/teenused/isikut-toendavad-dokumendid/mobiil-id/") );
		break;
	case PagePin1Pin:
		d->changePin1Info->setCurrentWidget( d->changePin1InfoPin );
		d->changePin1PinpadInfo->setCurrentWidget( d->changePin1PinpadInfoPin );
		d->changePin1ValidateLabel->setText( tr("Current PIN1 code") );
		d->changePin1Validate->setFocus();
		d->b->setId( d->changePin1Change, PagePin1ChangePin );
		d->b->setId( d->changePin1PinpadChange, PagePin1ChangePin );
		break;
	case PagePin1Puk:
		d->changePin1Info->setCurrentWidget( d->changePin1InfoPuk );
		d->changePin1PinpadInfo->setCurrentWidget( d->changePin1PinpadInfoPuk );
		d->changePin1ValidateLabel->setText( tr("Current PUK code") );
		d->changePin1Validate->setFocus();
		d->b->setId( d->changePin1Change, PagePin1ChangePuk );
		d->b->setId( d->changePin1PinpadChange, PagePin1ChangePuk );
		break;
	case PagePin1Unblock:
		d->changePin1Info->setCurrentWidget( d->changePin1InfoUnblock );
		d->changePin1PinpadInfo->setCurrentWidget( d->changePin1PinpadInfoUnblock );
		d->changePin1ValidateLabel->setText( tr("Current PUK code") );
		d->changePin1Validate->setFocus();
		d->b->setId( d->changePin1Change, PagePin1ChangeUnblock );
		d->b->setId( d->changePin1PinpadChange, PagePin1ChangeUnblock );
		break;
	case PagePin1ChangePin:
		if( !t.isPinpad() && !d->validatePin( QSmartCardData::Pin1Type, false,
				d->changePin1Validate->text(), d->changePin1New->text(), d->changePin1Repeat->text() ) )
		{
			d->clearPins();
			break;
		}
		if( t.isPinpad() )
			d->showLoading( tr("Enter PIN/PUK codes on PinPad") );
		else
			d->showLoading( tr("Changing %1 code").arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
		if( d->validateCardError( QSmartCardData::Pin1Type, 1024,
				d->smartcard->change( QSmartCardData::Pin1Type, d->changePin1New->text(), d->changePin1Validate->text() ) ) )
		{
			QMessageBox::information( this, windowTitle(), tr("%1 changed!").arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
			setDataPage( PageCert );
		}
		d->clearPins();
		break;
	case PagePin1ChangePuk:
	case PagePin1ChangeUnblock:
		if( !t.isPinpad() && !d->validatePin( QSmartCardData::Pin1Type, true,
				d->changePin1Validate->text(), d->changePin1New->text(), d->changePin1Repeat->text() ) )
		{
			d->clearPins();
			break;
		}

		if( !t.isPinpad() )
		{
			if( index == PagePin1ChangePuk )
				d->showLoading( tr("Changing %1 code")	.arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
			else
				d->showLoading( tr("Unblocking %1 code").arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
		}
		else
			d->showLoading( tr("Enter PIN/PUK codes on PinPad") );

		if( d->validateCardError( QSmartCardData::Pin1Type, 1025,
				d->smartcard->unblock( QSmartCardData::Pin1Type, d->changePin1New->text(), d->changePin1Validate->text() ) ) )
		{
			if( index == PagePin1ChangePuk )
				QMessageBox::information( this, windowTitle(), tr("%1 changed!")
					.arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
			else
				QMessageBox::information( this, windowTitle(), tr("%1 changed and yours current sertificates blocking has been removed!")
					.arg( QSmartCardData::typeString( QSmartCardData::Pin1Type ) ) );
			updateData();
			setDataPage( PageCert );
		}
		d->clearPins();
		break;
	case PagePin2Pin:
		d->changePin2Info->setCurrentWidget( d->changePin2InfoPin );
		d->changePin2PinpadInfo->setCurrentWidget( d->changePin2PinpadInfoPin );
		d->changePin2ValidateLabel->setText( tr("Current PIN2 code") );
		d->changePin2Validate->setFocus();
		d->b->setId( d->changePin2Change, PagePin2ChangePin );
		d->b->setId( d->changePin2PinpadChange, PagePin2ChangePin );
		break;
	case PagePin2Puk:
		d->changePin2Info->setCurrentWidget( d->changePin2InfoPuk );
		d->changePin2PinpadInfo->setCurrentWidget( d->changePin2PinpadInfoPuk );
		d->changePin2ValidateLabel->setText( tr("Current PUK code") );
		d->changePin2Validate->setFocus();
		d->b->setId( d->changePin2Change, PagePin2ChangePuk );
		d->b->setId( d->changePin2PinpadChange, PagePin2ChangePuk );
		break;
	case PagePin2Unblock:
		d->changePin2Info->setCurrentWidget( d->changePin2InfoUnblock );
		d->changePin2PinpadInfo->setCurrentWidget( d->changePin2PinpadInfoUnblock );
		d->changePin2ValidateLabel->setText( tr("Current PUK code") );
		d->changePin2Validate->setFocus();
		d->b->setId( d->changePin2Change, PagePin2ChangeUnblock );
		d->b->setId( d->changePin2PinpadChange, PagePin2ChangeUnblock );
		break;
	case PagePin2ChangePin:
		if( !t.isPinpad() && !d->validatePin( QSmartCardData::Pin2Type, false,
				d->changePin2Validate->text(), d->changePin2New->text(), d->changePin2Repeat->text() ) )
		{
			d->clearPins();
			break;
		}
		if( t.isPinpad() )
			d->showLoading( tr("Enter PIN/PUK codes on PinPad") );
		else
			d->showLoading( tr("Changing %1 code").arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
		if( d->validateCardError( QSmartCardData::Pin2Type, 1024,
				d->smartcard->change( QSmartCardData::Pin2Type, d->changePin2New->text(), d->changePin2Validate->text() ) ) )
		{
			QMessageBox::information( this, windowTitle(), tr("%1 changed!").arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
			setDataPage( PageCert );
		}
		d->clearPins();
		break;
	case PagePin2ChangePuk:
	case PagePin2ChangeUnblock:
		if( !t.isPinpad() && !d->validatePin( QSmartCardData::Pin2Type, true,
				d->changePin2Validate->text(), d->changePin2New->text(), d->changePin2Repeat->text() ) )
		{
			d->clearPins();
			break;
		}

		if( !t.isPinpad() )
		{
			if( index == PagePin2ChangePuk )
				d->showLoading( tr("Changing %1 code").arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
			else
				d->showLoading( tr("Unblocking %1 code").arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
		}
		else
			d->showLoading( tr("Enter PIN/PUK codes on PinPad") );

		if( d->validateCardError( QSmartCardData::Pin2Type, 1025,
				d->smartcard->unblock( QSmartCardData::Pin2Type, d->changePin2New->text(), d->changePin2Validate->text() ) ) )
		{
			if( index == PagePin2ChangePuk )
				QMessageBox::information( this, windowTitle(), tr("%1 changed!")
					.arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
			else
				QMessageBox::information( this, windowTitle(), tr("%1 changed and yours current sertificates blocking has been removed!")
					.arg( QSmartCardData::typeString( QSmartCardData::Pin2Type ) ) );
			updateData();
			setDataPage( PageCert );
		}
		d->clearPins();
		break;
	case PagePuk:
		d->changePukValidate->setFocus();
		break;
	case PagePukChange:
		if( !t.isPinpad() && !d->validatePin( QSmartCardData::PukType, false,
				d->changePukValidate->text(), d->changePukNew->text(), d->changePukRepeat->text() ) )
		{
			d->clearPins();
			break;
		}
		if( t.isPinpad() )
			d->showLoading( tr("Enter PIN/PUK codes on PinPad") );
		else
			d->showLoading( tr("Changing %1 code").arg( QSmartCardData::typeString( QSmartCardData::PukType ) ) );
		if( d->validateCardError( QSmartCardData::PukType, 1024,
				d->smartcard->change( QSmartCardData::PukType, d->changePukNew->text(), d->changePukValidate->text() ) ) )
		{
			QMessageBox::information( this, windowTitle(), tr("%1 changed!").arg( QSmartCardData::typeString( QSmartCardData::PukType ) ) );
			setDataPage( PageCert );
		}
		d->clearPins();
		break;
	default: break;
	}
	d->hideLoading();
}

void MainWindow::showAbout()
{ (new AboutWidget( this ))->show(); }

void MainWindow::showDiagnostics()
{ (new DiagnosticsDialog( this ) )->open(); }

void MainWindow::showHelp()
{ QDesktopServices::openUrl( Common::helpUrl() ); }

void MainWindow::showSettings()
{
#if defined(Q_OS_WIN) || defined(Q_OS_MAC)
	(new SettingsDialog( this ) )->open();
#endif
}

void MainWindow::showWarning( const QString &msg )
{ QMessageBox::warning( this, windowTitle(), msg ); }

QSmartCard* MainWindow::smartCard() const { return d->smartcard; }

void MainWindow::updateData()
{
	d->hideLoading();
	QSmartCardData t = d->smartcard->data();
	d->cardInfo->setHidden( !t.card().isEmpty() && t.authCert().isNull() );

	if( !t.authCert().isNull() )
	{
		d->cardInfo->setCurrentWidget( d->cardInfoPage );

		d->changePin1Stack->setCurrentIndex( t.isPinpad() );
		d->changePin2Stack->setCurrentIndex( t.isPinpad() );
		d->changePukStack->setCurrentIndex( t.isPinpad() );

		d->loadPicture->setText( tr("<a href=\"loadPicture\">Load picture</a>") );
		d->frameBirth->setHidden( t.cardType() != QSmartCardData::IDCard );
		d->frameCitizen->setHidden( t.cardType() != QSmartCardData::IDCard );
		d->frameRegNr->setHidden( t.cardType() != QSmartCardData::Tempel );
		d->frameEmail->setHidden( t.cardType() == QSmartCardData::Tempel );

		d->personalName->setText( t.data( QSmartCardData::FullName ).toString() );
		d->personalRegNr->setText( t.data( QSmartCardData::Id ).toString() );
		d->personalCode->setText( t.data( QSmartCardData::Id ).toString() );
		d->personalBirth->setText( QString( "%1 %2" )
			.arg( SslCertificate::formatDate( t.data( QSmartCardData::BirthDate ).toDateTime(), "dd. MMMM yyyy" ) )
			.arg( t.data( QSmartCardData::BirthPlace ).toString().isEmpty() ? "" : ", " + t.data( QSmartCardData::BirthPlace ).toString() ) );
		d->personalCitizen->setText( t.data( QSmartCardData::Citizen ).toString() );
		d->personalEmail->setText( t.data( QSmartCardData::Email ).toString() );

		d->frameValidTill->setHidden( t.cardType() == QSmartCardData::DigiID );
		d->labelInValid->setHidden( t.isValid() );
		d->labelValid->setHidden( !t.isValid() );
		d->labelValidTillDate->setText( SslCertificate::formatDate( t.data( QSmartCardData::Expiry ).toDateTime(), "dd. MMMM yyyy" ) );
		d->labelNewCard->setVisible( !t.isValid() && t.cardType() != QSmartCardData::DigiID );
		d->frameValidDocument->setHidden( t.cardType() == QSmartCardData::DigiID );
		d->labelDigiID->setHidden( t.cardType() != QSmartCardData::DigiID );
		d->labelCardNo->setText( t.data( QSmartCardData::DocumentId ).toString() );

		d->authTill->setText( SslCertificate::formatDate( t.authCert().expiryDate().toLocalTime(), "dd. MMMM yyyy" ) );
		d->signTill->setText( SslCertificate::formatDate( t.signCert().expiryDate().toLocalTime(), "dd. MMMM yyyy" ) );

		d->authValidity->setHidden( t.retryCount( QSmartCardData::Pin1Type ) == 0 );
		d->authInValidity->setHidden( t.retryCount( QSmartCardData::Pin1Type ) != 0 );
		d->signValidity->setHidden( t.retryCount( QSmartCardData::Pin2Type ) == 0 );
		d->signInValidity->setHidden( t.retryCount( QSmartCardData::Pin2Type ) != 0 );
		if ( t.retryCount( QSmartCardData::Pin1Type ) == 0 )
			d->authInValidity->setText( t.authCert().isValid() ? tr( "valid but blocked" ) : tr( "invalid and blocked" ) );
		else
			d->authValidity->setText( t.authCert().isValid() ? tr( "valid and applicable" ) : tr( "expired" ) );
		if ( t.retryCount( QSmartCardData::Pin2Type ) == 0 )
			d->signInValidity->setText( t.signCert().isValid() ? tr( "valid but blocked" ) : tr( "invalid and blocked" ) );
		else
			d->signValidity->setText( t.signCert().isValid() ? tr( "valid and applicable" ) : tr( "expired" ) );

		d->authUsageCount->setText( tr( "Authentication key has been used %1 times" ).arg( t.usageCount( QSmartCardData::Pin1Type ) ) );
		d->signUsageCount->setText( tr( "Signature key has been used %1 times" ).arg( t.usageCount( QSmartCardData::Pin2Type ) ) );
		d->authUsageCount->setHidden( t.retryCount( QSmartCardData::Pin1Type ) == 0 );
		d->signUsageCount->setHidden( t.retryCount( QSmartCardData::Pin2Type ) == 0 );

		int authDays = qMax( 0, QDateTime::currentDateTime().daysTo( t.authCert().expiryDate().toLocalTime() ) );
		int signDays = qMax( 0, QDateTime::currentDateTime().daysTo( t.signCert().expiryDate().toLocalTime() ) );
		bool authExpires = ( authDays >= 0 && authDays <= 105 && t.retryCount( QSmartCardData::Pin1Type ) != 0 );
		bool signExpires = ( signDays >= 0 && signDays <= 105 && t.retryCount( QSmartCardData::Pin2Type ) != 0 );
		d->authCertExpired->setText( authExpires ? ( t.authCert().isValid() ?
											tr( "Certificate will expire in %1 days" ).arg( authDays ) :
											tr( "Certificate is expired" ) ) : "" );
		d->signCertExpired->setText( signExpires ? ( t.signCert().isValid() ?
											tr( "Certificate will expire in %1 days" ).arg( signDays ) :
											tr( "Certificate is expired" ) ) : "" );
		d->authChangePin->setVisible( t.retryCount( QSmartCardData::Pin1Type ) > 0 );
		d->signChangePin->setVisible( t.retryCount( QSmartCardData::Pin2Type ) > 0 );
		d->authCertBlocked->setHidden( t.retryCount( QSmartCardData::Pin1Type ) > 0 );
		d->signCertBlocked->setHidden( t.retryCount( QSmartCardData::Pin2Type ) > 0 );
		d->authRevoke->setVisible( t.retryCount( QSmartCardData::Pin1Type ) == 0 && t.retryCount( QSmartCardData::PukType ) > 0 );
		d->signRevoke->setVisible( t.retryCount( QSmartCardData::Pin2Type ) == 0 && t.retryCount( QSmartCardData::PukType ) > 0 );

		d->updateCert->setVisible(
			t.cardType() == QSmartCardData::IDCard &&
			t.version() != QSmartCardData::Ver_3_0 &&
			( authDays <= 105 || signDays <= 105 ) &&
			t.retryCount( QSmartCardData::Pin1Type ) > 0 );

		d->pukLocked->setVisible( t.retryCount( QSmartCardData::PukType ) == 0 );
		d->pukChange->setVisible( t.retryCount( QSmartCardData::PukType ) > 0 );
		d->pukLink1->setVisible( t.retryCount( QSmartCardData::PukType ) > 0 );
		d->pukLink2->setVisible( t.retryCount( QSmartCardData::PukType ) > 0 );

		if( d->dataWidget->currentIndex() == PageEmpty )
			setDataPage( t.retryCount( QSmartCardData::PukType ) == 0 ? PagePukInfo : PageCert );

#ifdef Q_OS_WIN
		CertStore store;
		if( !store.find( t.authCert() ) &&
			QMessageBox::question( this, tr( "Certificate store" ),
				tr( "Certificate is not registered in certificate store. Register now?" ),
				QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes ) == QMessageBox::Yes )
		{
			store.add( t.authCert(), t.card() );
			store.add( t.signCert(), t.card() );
		}
#endif
	}
	else
	{
		d->personalName->clear();
		d->personalRegNr->clear();
		d->personalCode->clear();
		d->personalBirth->clear();
		d->personalCitizen->clear();
		d->personalEmail->clear();
		d->loadPicture->setProperty( "PICTURE", QVariant() );
		d->loadPicture->clear();
		d->savePicture->hide();
		setDataPage( PageEmpty );

		if( !DiagnosticsDialog::isPCSCRunning() )
			d->cardInfo->setCurrentWidget( d->labelNoPCSC );
		else if( t.readers().isEmpty() )
			d->cardInfo->setCurrentWidget( d->labelNoReader );
		else if( t.cards().isEmpty() )
			d->cardInfo->setCurrentWidget( d->labelNoCard );
		else if( !t.card().isEmpty() )
			d->showLoading( tr( "Reading data" ) );
	}

	d->cards->clear();
	d->cards->addItems( t.cards() );
	d->cards->setVisible( t.cards().size() > 1 );
	d->cards->setCurrentIndex( d->cards->findText( t.card() ) );
}
